{*********************************************************}
{     TExcel Component Demo 2.0 for Delphi 1.0/2.0        }
{                                                         }
{     Copyright (c) 1996, 1997 by                         }
{                                                         }
{       Stefan Hoffmeister           (Advanced version)   }
{       Stefan.Hoffmeister@Uni-Passau.de                  }
{   and                                                   }
{       Tibor F. Liska                (Basic version)     }
{       Tel/Fax:    00-36-1-165-2019                      }
{       Office:     00-36-1-269-8284                      }
{       E-mail: liska@sztaki.hu                           }
{                                                         }
{*********************************************************}
unit Excels;
(*{$D-} {$L-} {$Q-} {$R-} {$S-}
*)
interface

uses
  Forms, Classes, DdeMan, SysUtils;

type
  TCellDir = (dirNone, dirUp, dirDown, dirLeft, dirRight);
  TNewSheet = ( FromTemplate,
                FromActiveSheet,
                Sheet,
                ChartSelection,
                Macro,
                IntMacro,
                Reserved,
                VBModule,
                Dialog);


  ExcelError = class(Exception);

  TExcelRow = Integer;
  TExcelCol = Integer;

  TCustomExcel = class(TComponent)
  private
      FDDEClientConv : TDdeClientConv;
      FExeName : TFileName;
      FProtocolsList,
      FEditEnvItemsList,
      FTopicsList,
      FFormatsList: TStringList;
      FConnected : Boolean;
      FOnClose: TNotifyEvent;
      FOnOpen: TNotifyEvent;
    procedure SetExeName(const Value: TFileName);
    procedure SetConnect(Value: Boolean);
    function GetReady: Boolean;
    function GetFormats: TStringList;
    function GetTopics: TStringList;
    function GetProtocols: TStringList;
    function GetEditEnvItems: TStringList;
    function GetSelection: string;
    procedure OpenLink(Sender: TObject);
    procedure ShutDown(Sender: TObject);
  protected
    function GetStrings(var List: TStringList; const Topic: string): TStringList;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Connect;
    procedure Disconnect;
    function Request(const Item: string): string;
    procedure Exec(const Cmd: string);
    procedure LocateExcel; virtual;
    property Connected: Boolean read FConnected write SetConnect;
    property DDEConv: TDdeCLientConv read FDDEClientConv;
    property Ready: Boolean read GetReady;
    property Formats: TStringList read GetFormats;
    property Topics: TStringList read GetTopics;
    property Protocols: TStringList read GetProtocols;
    property EditEnvItems: TStringList read GetEditEnvItems;
    property Selection: string read GetSelection;
    property ExeName: TFileName read FExeName write SetExeName;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
  end;

  TExcel = class(TCustomExcel)
  private
      FMoveActiveCell : TCellDir;
      FDecimals: Word;
      FDecimalSeparator: Char;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AutoMoveActiveCell;
    procedure Select (Row: TExcelRow; Col: TExcelCol);
    procedure Move(deltaRow, deltaCol: Integer);
    procedure Insert(const s: string);
    procedure InsertAt(Row: TExcelRow; Col: TExcelCol; const s: string);
    procedure PutExt(e: Extended);
    procedure PutExtAt(Row: TExcelRow; Col: TExcelCol; e: Extended);
    procedure PutStr(const s: string);
    procedure PutStrAt(Row: TExcelRow; Col: TExcelCol; const s: string);
  published
    property MoveActiveCell: TCellDir read FMoveActiveCell write FMoveActiveCell default dirDown;
    property ExeName;
    property Decimals: Word read FDecimals write FDecimals default 0;
    property DecimalSeparator: Char read FDecimalSeparator write FDecimalSeparator default '.';
    property OnClose;
    property OnOpen;
  end;


  TAdvExcel = class(TExcel)
  private
      FMacro : TFileName;
  public
    procedure OpenMacroFile(const Fn: TFileName);
    procedure CloseMacroFile;
    procedure RunMacro(const Mn: string);
    procedure StartTable;
    procedure EndTable;
    procedure NewSheet(SheetType: TNewSheet; const TemplateName: string);
    procedure NewWorkbook(SheetType: TNewSheet; const TemplateName: string);
    procedure EchoOn;
    procedure EchoOff;
    procedure DisableInput;
    procedure EnableInput;
    procedure PutInt(i: Longint);
    procedure PutIntAt(Row: TExcelRow; Col: TExcelCol; i: Longint);
    procedure PutDate(d: TDateTime);
    procedure PutDateAt(Row: TExcelRow; Col: TExcelCol; d: TDateTime);
    procedure PutTime(d: TDateTime);
    procedure PutTimeAt(Row: TExcelRow; Col: TExcelCol; d: TDateTime);
    procedure PutDateTime(d: TDateTime);
    procedure PutDateTimeAt(Row: TExcelRow; Col: TExcelCol; d: TDateTime);
    procedure PutData(const AnArray: array of const);
    procedure PutDataAt(Row: TExcelRow; Col: TExcelCol; const AnArray: array of const;
                        FillDirection: TCellDir);
  end;

procedure Register;

implementation

{$IFDEF Win32}
   {$R EXCELS.D32}
{$ELSE}
   {$R EXCELS.D16}
{$ENDIF}

uses
{$IFDEF Win32}
   Windows,
{$ENDIF}
   ShellApi;

procedure Register;
begin
  RegisterComponents('Samples', [TAdvExcel, TExcel]);
end;

{ TCustomExcel }

constructor TCustomExcel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  if not (csDesigning in ComponentState) then
  begin
    FDDEClientConv := TDdeClientConv.Create(nil);
    with FDDEClientConv do
    begin
      ConnectMode := ddeManual;
      OnOpen  := Self.OpenLink;
      OnClose := Self.ShutDown;
    end;
  end;

  SetExeName('Excel');
end;

destructor TCustomExcel.Destroy;
begin
  if Assigned(FFormatsList)      then FFormatsList.Free;
  if Assigned(FTopicsList)       then FTopicsList.Free;
  if Assigned(FProtocolsList)    then FProtocolsList.Free;
  if Assigned(FEditEnvItemsList) then FEditEnvItemsList.Free;

  if not (csDesigning in ComponentState) then
    FDDEClientConv.Free;

  inherited Destroy;
end;

procedure TCustomExcel.SetExeName(const Value: TFileName);
begin
  Disconnect;

  FExeName := ChangeFileExt(Value, '');
  if not (csDesigning in ComponentState) then
    FDDEClientConv.ServiceApplication := FExeName;
end;

procedure TCustomExcel.SetConnect(Value: Boolean);
begin
  if FConnected = Value then Exit;
  if Value then
    Connect
  else
    Disconnect;
end;

function TCustomExcel.GetReady: Boolean;
begin
  Application.ProcessMessages;
  Result := 'Ready' = Request('Status');
end;

function TCustomExcel.GetFormats: TStringList;
begin
  Result := GetStrings(FFormatsList, 'Formats');
end;

function TCustomExcel.GetTopics: TStringList;
begin
  Result := GetStrings(FTopicsList, 'Topics');
end;

function TCustomExcel.GetProtocols: TStringList;
begin
  Result := GetStrings(FProtocolsList, 'Protocols');
end;

function TCustomExcel.GetEditEnvItems: TStringList;
begin
  Result := GetStrings(FEditEnvItemsList, 'EditEnvItems');
end;

function TCustomExcel.GetSelection: string;
begin
  Application.ProcessMessages;
  Result := Request('Selection');
end;

procedure TCustomExcel.OpenLink(Sender: TObject);
begin
  FConnected := True;
  if Assigned(FOnOpen) then FOnOpen(Self);
end;

procedure TCustomExcel.ShutDown(Sender: TObject);
begin
  FConnected := False;
  if Assigned(FOnClose) then FOnClose(Self);
end;

{ procedure relies on List <> nil }
procedure TABStringToStringList(TABString: PChar; List: TStringList);
  var
      StartPos,
      TabPos : PChar;
begin
  if TABString = nil then Exit;
  StartPos := TABString;
  TabPos := StrScan(StartPos, #9);
  while TabPos <> nil do
  begin
    TabPos[0] := #0; { replace #9 by #0 }
    List.Add(StrPas(StartPos)); { add format to list }
    StartPos := TabPos + 1; { position to after TAB}
    TabPos := StrScan(StartPos, #9);
  end;
  if StrLen(StartPos) > 0 then
    List.Add(StrPas(StartPos));
end;

function TCustomExcel.GetStrings(var List: TStringList;
                               const Topic: string): TStringList;
  var
      Reply : PChar;
begin
  Application.ProcessMessages;

  if not Assigned(List) then
    List := TStringList.Create
  else
    List.Clear;
  try
    Reply := FDDEClientConv.RequestData(Topic);
    { retrieve TAB-delimited list of formats }
    TABStringToStringList(Reply, List);
  finally
    StrDispose(Reply);
  end;

  Result := List;
end;

procedure TCustomExcel.Connect;
begin
  if not FConnected then
  begin
    FDDEClientConv.SetLink('Excel', 'System');

    if FDDEClientConv.OpenLink then Exit;

    LocateExcel;
    if not FDDEClientConv.OpenLink then
      raise ExcelError.Create('Excel cannot start');
  end;
end;

procedure TCustomExcel.Disconnect;
begin
  if FConnected then
    FDDEClientConv.CloseLink;
end;

function TCustomExcel.Request(const Item: string): string;
  var
      Reply : PChar;
begin
  Reply := FDDEClientConv.RequestData(Item);
  if Reply = nil then
    Result := '*** No Reply ***'
  else
    Result := StrPas(Reply);
  StrDispose(Reply);
end;

procedure TCustomExcel.Exec(const Cmd: string);
  var
      Buffer : PChar;
begin
  Buffer := StrAlloc(Length(Cmd)+SizeOf(Char));
  try
    if not FDDEClientConv.ExecuteMacro(StrPCopy(Buffer, Cmd), False) then
    begin
      while FDDEClientConv.WaitStat or (not Ready) do
        Application.ProcessMessages;   { Waiting for Excel }

      if not FDDEClientConv.ExecuteMacro(Buffer, True) then
        raise ExcelError.Create('"'+Cmd+'" not accepted by Excel');
    end;
  finally
    StrDispose(Buffer);
  end;
end;

procedure TCustomExcel.LocateExcel;
  const
      BufferSize = {$IFDEF Win32} 540 {$ELSE} 80 {$ENDIF};
  var
      Buffer : PChar;
      StringPosition : PChar;
      ReturnedData: Longint;
begin
  Buffer := StrAlloc(BufferSize);
  try
    { get the first entry, don't bother about the version !}
    ReturnedData := BufferSize;
    StrPCopy(Buffer, '.xls');
    RegQueryValue(hKey_Classes_Root, Buffer, Buffer, ReturnedData);
    if StrLen(Buffer) > 0 then
    begin
      StrCat(Buffer, '\shell\Open\command');
      ReturnedData := BufferSize;
      if RegQueryValue(hKey_Classes_Root, Buffer, Buffer, ReturnedData) = ERROR_SUCCESS then
      begin
        StringPosition := StrUpper(Buffer);
        while StrPos(StringPosition+1, 'EXCEL.EXE') <> nil do
          StringPosition := StrPos(StringPosition+1, 'EXCEL.EXE');
        StrCopy(StringPosition + Length('EXCEL.EXE'), ''); { cut off string }
        if FileExists(StrPas(Buffer)) then
          ExeName := StrPas(Buffer);
      end;
    end;
  finally
    StrDispose(Buffer);
  end;
end;

{ TExcel }

constructor TExcel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMoveActiveCell := dirDown;
  FDecimalSeparator := '.';
end;

procedure TExcel.AutoMoveActiveCell;
begin
  case FMoveActiveCell of
    dirNone: ;
    dirUp:    Exec('[SELECT("R[-1]C")]');
    dirDown:  Exec('[SELECT("R[1]C")]');
    dirLeft:  Exec('[SELECT("RC[-1]")]');
    dirRight: Exec('[SELECT("RC[1]")]');
  end;
end;

procedure TExcel.Select(Row: TExcelRow; Col: TExcelCol);
begin
  if (Row <> 0) and (Col <> 0) then
    Exec(Format('[SELECT("R%dC%d")]', [Row, Col]));
end;

procedure TExcel.Move(deltaRow, deltaCol: Integer);
begin
  Exec(Format('[SELECT("R[%d]C[%d]")]', [deltaRow, deltaCol]));
end;

procedure TExcel.Insert(const s: string);
begin
  Exec(Format('[FORMULA(%s)]', [s]));
  AutoMoveActiveCell;
end;

procedure TExcel.InsertAt(Row: TExcelRow; Col: TExcelCol; const s: string);
begin
  if (Row <= 0) or (Col<=0) then
    Insert(s)
  else
    Exec(Format('[FORMULA(%s, "R%dC%d")]', [s, Row, Col]));
end;

procedure TExcel.PutExt(e: Extended);
begin
  PutExtAt(0, 0, e);
  AutoMoveActiveCell;
end;

procedure TExcel.PutExtAt(Row: TExcelRow; Col: TExcelCol; e: Extended);
  var
      SepPos : Integer;
      ExtString: string[30];
begin
  Str(e:0:FDecimals, ExtString);
  { this will always return an "American" style number }
  if FDecimalSeparator <> '.' then
  begin
    SepPos := Pos('.', ExtString);
    if SepPos > 0 then
      ExtString[SepPos] := FDecimalSeparator;
  end;
  InsertAt(Row, Col, ExtString);
end;

procedure TExcel.PutStr(const s: string);
begin
  PutStrAt(0, 0, s);
end;

procedure TExcel.PutStrAt(Row: TExcelRow; Col: TExcelCol; const s: string);
begin
  InsertAt(Row, Col, Format('"%s"', [s]));
end;

{ TAdvExcel }

procedure TAdvExcel.OpenMacroFile(const Fn: TFileName);
begin
  if FMacro <> UpperCase(ExtractFileName(Fn)) then
  begin
    if FMacro <> '' then
      CloseMacroFile;
    Exec(Format('[OPEN("%s")]', [Fn]));
    Exec('[HIDE()]');
    FMacro := UpperCase(ExtractFileName(Fn));
  end;
end;

procedure TAdvExcel.CloseMacroFile;
begin
  if FMacro <> '' then
  begin
    Exec('[UNHIDE("' + FMacro + '")]');
    Exec('[CLOSE(FALSE)]');
    FMacro := '';
  end;
end;

procedure TAdvExcel.RunMacro(const Mn: string);
begin
  if FMacro = '' then
    raise Exception.Create('No open macro');
  Exec(Format('[RUN("%s!%S";FALSE)]', [FMacro, Mn]));
end;

procedure TAdvExcel.StartTable;
begin
  Exec('[APP.MINIMIZE()]');
  Exec('[NEW(1)]');
  PutStrAt(1, 1, 'Table not ready');
end;

procedure TAdvExcel.EndTable;
begin
  PutStrAt(1, 1, '');
  Exec('[APP.RESTORE()]');
end;

procedure TAdvExcel.NewSheet(SheetType: TNewSheet; const TemplateName: string);
  var
      x : Integer;
begin
  x := ord(SheetType);
  if SheetType = Reserved then
    raise Exception.Create('Use of reserved type');
  if SheetType = FromTemplate then
    Exec(Format('[WORKBOOK.INSERT("%s")]', [TemplateName]))
  else
    Exec(Format('[WORKBOOK.INSERT(%d)]', [ord(SheetType)-1]));
end;

procedure TAdvExcel.NewWorkbook(SheetType: TNewSheet; const TemplateName: string);
begin
  if SheetType = FromTemplate then
    Exec(Format('[NEW("%s")]', [TemplateName]))
  else
    Exec(Format('[NEW(%d)]', [ord(SheetType)-1]));
end;

procedure TAdvExcel.EchoOn;
begin
  Exec('[ECHO(TRUE)]');
end;

procedure TAdvExcel.EchoOff;
begin
  Exec('[ECHO(FALSE)]');
end;

procedure TAdvExcel.DisableInput;
begin
  Exec('[DISABLE.INPUT(TRUE)]');
end;

procedure TAdvExcel.EnableInput;
begin
  Exec('[DISABLE.INPUT(FALSE)]');
end;

procedure TAdvExcel.PutInt(i: Longint);
begin
  PutIntAt(0, 0, i);
end;

procedure TAdvExcel.PutIntAt(Row: TExcelRow; Col: TExcelCol; i: Longint);
begin
  InsertAt(Row, Col, IntToStr(i));
end;

procedure TAdvExcel.PutDate(d: TDateTime);
begin
  PutDateAt(0, 0, d);
end;

procedure TAdvExcel.PutDateAt(Row: TExcelRow; Col: TExcelCol; d: TDateTime);
begin
  PutStrAt(Row, Col, DateToStr(d));
end;

procedure TAdvExcel.PutTime(d: TDateTime);
begin
  PutTimeAt(0, 0, d);
end;

procedure TAdvExcel.PutTimeAt(Row: TExcelRow; Col: TExcelCol; d: TDateTime);
begin
  PutStrAt(Row, Col, TimeToStr(d));
end;

procedure TAdvExcel.PutDateTime(d: TDateTime);
begin
  PutDateTimeAt(0, 0, d);
end;

procedure TAdvExcel.PutDateTimeAt(Row: TExcelRow; Col: TExcelCol; d: TDateTime);
begin
  PutStrAt(Row, Col, DateTimeToStr(d));
end;

procedure TAdvExcel.PutData(const AnArray: array of const);
begin
  PutDataAt(0, 0, AnArray, MoveActiveCell);
end;

procedure TAdvExcel.PutDataAt( Row: TExcelRow; Col: TExcelCol;
                               const AnArray: array of const;
                               FillDirection: TCellDir);
  var
      i: Integer;
      AutoMove: TCellDir;
begin
  Select(Row, Col);
  Row := 0; { invalidate location }
  Col := 0;
  if (FillDirection = dirNone) and (High(AnArray)>0) then
    raise ExcelError.Create('Putting an array needs auto-moving cell!');
  AutoMove := Self.MoveActiveCell;
  Self.MoveActiveCell := FillDirection;
  try
    for i:= Low(AnArray) to High(AnArray) do
    with AnArray[i] do
      case VType of
        vtExtended: PutExtAt(0, 0, VExtended^);
        vtInteger:  PutIntAt(0, 0, VInteger);
        vtChar:     PutStrAt(0, 0, VChar);
        vtString:   PutStrAt(0, 0, VString^);
        vtPChar:    PutStrAt(0, 0, StrPas(VPChar));
      {$IFDEF Win32}
        vtAnsiString: PutStrAt(0, 0, StrPas(VPChar));
      {$ENDIF}
      else
        raise ExcelError.Create('Data type not supported');
      end; { case }
  finally
    Self.MoveActiveCell := AutoMove;
  end;
end;

end.
